<?php

namespace Drupal\webform_email_reply_threads\Form;

use Drupal\awareness\Entity\EntityTypeManagerAwareTrait;
use Drupal\awareness\Routing\RouteMatchAwareTrait;
use Drupal\awareness\Session\CurrentUserAwareTrait;
use Drupal\Core\Form\FormStateInterface;
use Drupal\webform_email_reply_threads\Service\ReplySenderAwareTrait;
use Drupal\webform_email_reply_threads\Service\ThreadTracking;
use Drupal\webform_email_reply_threads\Service\ThreadTrackingAwareTrait;
use Drupal\webform_email_reply_threads\Util\EmailAddressFormatter;
use Drupal\webform_email_reply_threads\WebformEmailReplyThreadInterface;
use Email\Parse;

/**
 * Trait for common webform email reply functionality.
 */
trait WebformEmailReplyThreadReplyFormTrait {

  use CurrentUserAwareTrait;
  use EntityTypeManagerAwareTrait;
  use ReplySenderAwareTrait;
  use RouteMatchAwareTrait;
  use ThreadTrackingAwareTrait;

  /**
   * Builds out the reply form.
   *
   * @return array
   *   The form element with the reply form build it.
   */
  protected function buildReplyForm(WebformEmailReplyThreadInterface $thread) {
    $uuid = !$thread->isNew() ? $this->getRouteMatch()->getParameter('uuid') : NULL;
    $current_recipient = $this->getThreadTracking()->getCurrentThreadRecipient($thread, $uuid);

    $default_from_email = NULL;
    if ($current_recipient) {
      $default_from_email = EmailAddressFormatter::formatArray($current_recipient);
    }
    else {
      // No current recipient, use logged-in user.
      $account = $this->getCurrentUser();
      if ($account->isAuthenticated()) {
        $default_from_email = EmailAddressFormatter::format($this->getCurrentUser()->getEmail(), $this->getCurrentUser()->getDisplayName());
      }
    }

    $form['#thread'] = $thread;
    $form['#tree'] = TRUE;

    $recipients = $this->getThreadTracking()->getThreadRecipients($thread);

    $form['from_address'] = [
      '#type' => 'textfield',
      '#title' => t('From'),
      '#default_value' => $default_from_email,
      '#description' => t('The email address to send from.'),
      '#required' => TRUE,
      '#disabled' => $current_recipient,
      '#element_validate' => [
        [$this, 'validateParsedEmail'],
      ],
    ];

    $default_recipients = $recipients;
    if ($current_recipient) {
      $default_recipients = array_filter($default_recipients, function ($recipient) use ($current_recipient) {
        return $recipient['id'] != $current_recipient['id'];
      });
    }
    $default_recipients = array_map(function ($recipient) {
      return EmailAddressFormatter::formatArray($recipient);
    }, $default_recipients);
    $default_recipients = implode(', ', $default_recipients);

    $form['email'] = [
      '#type' => 'textfield',
      '#title' => t('To'),
      '#description' => t('The email address(es) to send to. Multiple emails should be separated by a comma, with no spaces.'),
      '#required' => TRUE,
      '#mulitple' => TRUE,
      '#message_key' => $thread->isNew() ? 'new_thread' : 'thread_reply',
      '#element_validate' => [
        [$this, 'validateParsedEmail'],
      ],

      '#default_value' => $default_recipients,
      '#disabled' => !$thread->isNew() && !$this->getCurrentUser()->hasPermission('administer webform_email_reply_threads'),
    ];

    $form['message'] = [
      '#type' => 'webform_html_editor',
      '#title' => t('Message'),
      '#required' => TRUE,
    ];

    $form['attachment'] = [
      '#type' => 'managed_file',
      '#title' => t("Attachment"),
      '#upload_location' => 'public://webform_email_reply/',
    ];

    return $form;
  }

  /**
   * Element validation callback for parsed email addresses.
   */
  public static function validateParsedEmail(array $element, FormStateInterface $form_state) {
    $value = $form_state->getValue($element['#parents']);
    $result = Parse::getInstance()->parse($value);

    if (!$result['success']) {
      $form_state->setErrorByName($element['#name'], $result['reason']);
      return;
    }

    // If required, ensure we've got at least one valid address.
    if ($element['#required'] && count($result['email_addresses']) < 1) {
      $params = ['@name' => $element['#title']];
      if (!empty($element['#multiple'])) {
        $form_state->setErrorByName($element['#name'], t('One or more valid email addresses are required for @name.', $params));
      }
      else {
        $form_state->setErrorByName($element['#name'], t('A valid email addresses is required for @name.', $params));
      }
      return;
    }

    // If not multiple, ensure we've only got a single value.
    if (!empty($element['#multiple']) && count($result['email_addresses']) > 1) {
      $params = ['@name' => $element['#title']];
      $form_state->setErrorByName($element['#name'], t('Only one email address allowed for @name.', $params));
    }

    // Allow message key to be passed in.
    if (!empty($element['#message_key'])) {
      $result['message_key'] = $element['#message_key'];
    }

    $form_state->set($element['#parents'], $result);
  }

  /**
   * Send a reply to a thread from a reply form.
   *
   * @param array $element
   *   Form element representing the reply form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   */
  protected function sendReply(array $element, FormStateInterface $form_state) {
    /** @var \Drupal\webform_email_reply_threads\WebformEmailReplyThreadInterface $thread */
    $thread = $element['#thread'];

    // Get the sender.
    $from = $form_state->get(array_merge($element['#parents'], ['from_address', 'email_addresses', 0]));

    // Get the recipients.
    $recipients = $form_state->get(array_merge($element['#parents'], ['email']));
    $message_key = $recipients['message_key'];
    $recipients = $recipients['email_addresses'];

    // Handle recipients.
    if ($message_key == 'new_thread') {
      $sender_type = ThreadTracking::ORIGINAL_SENDER;
      $recipient_type = ThreadTracking::ORIGINAL_RECIPIENT;
    }
    else {
      $sender_type = ThreadTracking::OTHER_RECIPIENT;
      $recipient_type = ThreadTracking::OTHER_RECIPIENT;
    }
    $uid = $this->getCurrentUser()->isAuthenticated() ? $this->getCurrentUser()->id() : NULL;
    $sender_id = $this->getThreadTracking()->addThreadRecipient($thread, $sender_type, $uid, $from['simple_address'], $from['name_parsed'] ?: NULL);
    foreach ($recipients as $recipient) {
      $this->getThreadTracking()->addThreadRecipient($thread, $recipient_type, NULL, $recipient['simple_address'], $recipient['name_parsed'] ?: NULL);
    }

    // Get the attachment.
    $file = NULL;
    $fid = $form_state->getValue(array_merge($element['#parents'], ['attachment']));
    if ($fid = reset($fid)) {
      /** @var \Drupal\file\FileInterface $file */
      $file = $this->getEntityTypeManager()
        ->getStorage('file')
        ->load($fid);
    }

    // Send it.
    $this->getReplySender()->sendReply(
      $thread,
      $message_key,
      $sender_id,
      $form_state->getValue(array_merge($element['#parents'], ['message'])),
      $file
    );
  }

}
